==> indsl-4.0.2.tar.gz: indsl-4.0.2/indsl/ts_utils/ts_utils.py <==
# Copyright 2021 Cognite AS
import warnings

from typing import Callable, List, Union

import numpy as np
import pandas as pd
import scipy.integrate

from cognite.client.utils._time import granularity_to_ms

from ..exceptions import UserValueError


def get_fixed_freq(start_time, end_time, granularity):
    """Create a Timeseries with a fixed frequency.

    Args:
        start_time (Union(pandas.Datetime,str, int, float)): Start time value to be converted to Timestamp.
        end_time (Union(pandas.Datetime,str, int, float)): End time value to be converted to Timestamp.
        granularity (str): Offset which Timestamp will have. Must be in format <integer>(s|m|h|d). E.g. '5m', '3h' or '1d'.".

    Returns:
        numpy.ndarray: Timeseries with fixed frequency.
    """
    return np.arange(
        pd.Timestamp(start_time, unit="ms"),
        pd.Timestamp(end_time, unit="ms"),
        pd.Timedelta(milliseconds=granularity_to_ms(granularity)),
    )


def above_below(series: pd.Series, upper_limit: float, lower_limit: float):
    """Above/below
    Count the number of points above/below upper_limit/lower_limit.

    Args:
        series (pandas.Series): Data with a pandas.DatetimeIndex.
        upper_limit (float): Value for upper limit.
        lower_limit (float): Value for lower limit.

    Returns:
        dict: The numbers above/below upper_limit/lower_limit.
    """
    result = {}
    if upper_limit is not None:
        result["above_upper"] = (series > upper_limit).sum()

    if lower_limit is not None:
        result["below_lower"] = (series < lower_limit).sum()

    return result


def num_vals_in_boxes(series: pd.Series, num_boxes: int):
    """Count the number of points in each of the boxes between 0 and 100 %.

    Args:
        series (pandas.Series): Data with a pandas.DatetimeIndex.
        num_boxes (int): Number of boxes.

    Returns:
        dict: A list with the number of points in each box in keyword "data".
    """
    max_val = series.max()
    min_val = series.min()
    delta = (max_val - min_val) / float(num_boxes)
    eps = 1.0e-6

    boxes = []
    low_limit = min_val
    for idx in range(num_boxes):
        if idx == 0:
            num_value_in_box = ((series >= low_limit * (1 - eps)) & (series <= low_limit + delta * (1 + eps))).sum()
        else:
            num_value_in_box = ((series > low_limit) & (series <= low_limit + delta * (1 + eps))).sum()
        boxes.append(num_value_in_box)
        low_limit += delta

    return {"data": boxes}


def functional_mean(function: Callable, x_vals: List) -> np.ndarray:
    """Convenience method to calculate the mean of a function inbetween each x value.
    The last point is the function called at the last x value. This is required for the
    input and output lists to have the same lengths.

    Args:
        function (Callable): Function used to calculate the mean.
        x_vals (List): Data values used to calculate the mean of a function.

    Raises:
        RuntimeError: Checks that x_vals is not empty.

    Returns:
        numpy.ndarray: Array with the calculated values.
    """

    # Catch empty input
    if len(x_vals) == 0:
        raise RuntimeError("No data in the input timeseries.")

    y_vals = np.zeros(len(x_vals))
    for i in range(len(x_vals) - 1):
        y_vals[i] = scipy.integrate.quad(function, x_vals[i], x_vals[i + 1])[0] / (x_vals[i + 1] - x_vals[i])
    y_vals[len(x_vals) - 1] = function(x_vals[-1])
    return y_vals


def is_na_all(data: Union[pd.DataFrame, pd.Series]) -> bool:
    """Convenience method to test if all values in dataframe/series are NaN.

    Args:
        data (Union[pandas.DataFrame, pandas.Series]): Input data.

    Raises:
        UserValueError: Checks that the data type for the input data is pandas.DataFrame or pandas.Series.

    Returns:
        bool: True if all values are Nan, False if all values are not Nan.
    """
    if isinstance(data, pd.Series):
        return data.isnull().all()
    elif isinstance(data, pd.DataFrame):
        return pd.isna(data).all(axis=None)
    else:
        raise UserValueError("Convenience method only supports Series or DataFrame.")


def gaps_detector(timestamps: np.ndarray, threshold: int = 86400) -> np.ndarray:
    """Detect gaps bigger than a threshold in time series
    Taken from: https://github.com/cognitedata/data-quality-functions/blob/master/functions/completeness/function/completeness.py

    Args:
        timestamps (numpy.ndarray): Time array_like.
        threshold (int, optional): Threshold for gaps. Defaults to 86400 seconds, i.e one full day in epoch time.
    Returns:
        numpy.ndarray: 2D array of begin and end time of gaps in seconds epoch time.
    """
    if len(timestamps) == 0:
        return None

    mask = np.diff(timestamps) >= threshold
    gaps = np.column_stack([timestamps[:-1][mask], timestamps[1:][mask]])

    return gaps


def check_uniform(data: pd.DataFrame, unit: str = "ns") -> bool:
    """Convenience method to verify input time series is uniform.
    Args:
        data (pandas.DataFrame): Time series to check. Object must have a datetime-like index.
        unit (str, optional): Time unit of uniformity check. Follows Numpy DateTime Units convention. Defaults to "ns".
    Returns:
        bool: True if input timeseries is uniform, False if not uniform.
    """
    # Get time difference between successive observations
    timediff = np.diff(data.index)
    # Round to desired time unit (e.g. uniform on second scale)
    floored = timediff.astype(dtype=f"timedelta64[{unit}]")
    # Check whether uniform
    return np.all(floored[0] == floored[1:])


def make_uniform(data: pd.DataFrame, resolution: str = "1s", interpolation: str = None) -> pd.DataFrame:
    """Convenience method to make an input time series dataframe uniform per specified temporal resolution.
    Object must have a datetime-like index.
    Args:
        resolution (str):  Temporal resolution of uniform time series. Follows Pandas DateTime convention. Defaults to "1s".
        interpolation (str, optional): Method for optional interpolation. Follows pandas Resampler.interpolate. Defaults to None.
    Returns:
        pandas.DataFrame: Uniform Timeseries DataFrame.
    """
    # If no interpolation specified, do not perform interpolation
    if not interpolation:
        data = data.resample(resolution).mean()
    # Otherwise perform desired interpolation
    else:
        data = data.resample(rule=resolution).mean().interpolate(method=interpolation)
    return data


def time_parse(time_window: str = "60min", function_name: str = ""):
    """Check that time_window is given in a correct format.

    Args:
        time_window (str, optional): Time window input. Defaults to "60min".
            Accepted string format: '3w', '10d', '5h', '30min', '10s'.
            If the user gives a number without unit (such as '60'), it will be considered as the number of minutes.
        function_name (str, optional): Prints the name of the function that called time_parse in the error message if an error is raised. Defaults to "".

    Raises:
        UserValueError: Raise a value error if time_window is given in a wrong format.

    Returns:
        str: parsed time_window in string format.
    """
    if str.isdigit(time_window):  # if user gives '60' it will be considered as '60min'
        warnings.warn(
            f"Missing time unit in argument 'time_window' in {function_name} function, assuming {time_window}min.",
            UserWarning,
        )
        time_window = str(time_window) + "min"
    try:
        return pd.Timedelta(time_window)
    except ValueError:
        raise UserValueError(
            f"Time window should be a string in weeks, days, hours or minutes format:'3w', '10d', '5h', '30min', '10s' not {time_window}."
        ) from None


def time_to_points(data: pd.Series, time_window="60min"):
    """Find the average number of points given a time window.

    Args:
        data (pandas.Series): Data with a pandas.DatetimeIndex.
        time_window (str, optional): Length of the time period to compute the average of points. Defaults to "60min".
            Accepted string format: '3w', '10d', '5h', '30min', '10s'.

    Returns:
        int: average number of points in a time window.
    """
    time_window = time_parse(time_window)
    num_points = data.rolling(time_window).count().mean()

    return round(num_points)


def fill_gaps(
    data: pd.Series,
    granularity: str = None,
    ffill_resolution: str = None,
    interpolate_resolution: str = None,
    method: str = "ffill",
) -> pd.Series:
    """Mask nan/fill gaps bases on the time size of missing data

    Args:
        data (pandas.Series): Time series to impute.
        granularity (str, optional): Temporal resolution of uniform time series, before resampling. Defaults to None.
        ffill_resolution (str, optional): Gaps smaller than this will be filled with method. Defaults to None.
        interpolate_resolution (str, optional): Gaps smaller than this will be interpolated, larger than this will be filled by noise. Defaults to None.
        method (str, optional): Method used to fill nan. Defaults to 'ffill', but ‘backfill’, ‘bfill’ and ‘pad’ can also be used.
    Returns:
        pandas.DataFrame or pd.Series: Uniform, time series without nan values.
    """

    if not data.isnull().any():
        # no need to proceed
        return data

    if method not in ("backfill", "bfill", "pad", "ffill"):
        warnings.warn("Method should be in ('backfill', 'bfill', 'pad', 'ffill'). Set to ffill.", RuntimeWarning)
        method = "ffill"

    granularity_timedelata = pd.Timedelta("0T")
    try:
        granularity_timedelata = pd.Timedelta(granularity)
    except ValueError:
        raise ValueError(f"Unable to convert {granularity} to time resolution", RuntimeWarning)

    small_gap = 0
    if ffill_resolution:
        try:
            small_gap = granularity_timedelata / pd.Timedelta(ffill_resolution)
        except (ValueError, ZeroDivisionError):
            raise ValueError(f"Unable to convert {ffill_resolution} to time resolution", RuntimeWarning)

    inter_gap = None
    mask_gap = 0
    if interpolate_resolution is not None:
        try:
            mask_gap = granularity_timedelata / pd.Timedelta(interpolate_resolution)
            inter_gap = int(mask_gap)
        except (ValueError, ZeroDivisionError):
            raise ValueError(f"Unable to convert {interpolate_resolution} to time resolution", RuntimeWarning)

    # ffill all, after interpolate and adding noise
    # create groups based on size of gap
    nan_cumsum = (
        data.isnull().astype("int64").groupby(data.notnull().astype("int64").cumsum()).cumsum().mask(lambda x: x == 0)
    )
    helper = data.notnull()
    nan_group_size = nan_cumsum.groupby(helper.cumsum()).transform("max").mask(helper)

    # start with random filling for the big gaps
    mask = nan_group_size > inter_gap
    n = len(data[mask])
    data[mask] = np.random.normal(loc=np.mean(data), scale=np.std(data), size=n)

    # interpolate the inteprolate range
    mask = np.logical_and(nan_group_size > small_gap, nan_group_size <= mask_gap)
    data[mask] = data.interpolate(method="linear", limit_direction="forward", limit=inter_gap)

    return data.fillna(method=method)

==> indsl-4.1.1.tar.gz: indsl-4.1.1/indsl/ts_utils/ts_utils.py <==
# Copyright 2021 Cognite AS
import re
import warnings

from datetime import datetime
from typing import Callable, List, Union

import numpy as np
import pandas as pd
import scipy.integrate

from ..exceptions import UserValueError


_unit_in_ms_without_week = {"s": 1000, "m": 60000, "h": 3600000, "d": 86400000}
_unit_in_ms = {**_unit_in_ms_without_week, "w": 604800000}


def datetime_to_ms(dt):
    epoch = datetime.utcfromtimestamp(0)
    return int((dt - epoch).total_seconds() * 1000.0)


def time_string_to_ms(pattern, string, unit_in_ms):
    pattern = pattern.format("|".join(unit_in_ms))
    res = re.fullmatch(pattern, string)
    if res:
        magnitude = int(res.group(1))
        unit = res.group(2)
        return magnitude * unit_in_ms[unit]
    return None


def granularity_to_ms(granularity: str) -> int:
    ms = time_string_to_ms(r"(\d+)({})", granularity, _unit_in_ms_without_week)
    if ms is None:
        raise ValueError(
            "Invalid granularity format: `{}`. Must be on format <integer>(s|m|h|d). E.g. '5m', '3h' or '1d'.".format(
                granularity
            )
        )
    return ms


def time_ago_to_ms(time_ago_string: str) -> int:
    """Returns millisecond representation of time-ago string"""
    if time_ago_string == "now":
        return 0
    ms = time_string_to_ms(r"(\d+)({})-ago", time_ago_string, _unit_in_ms)
    if ms is None:
        raise ValueError(
            "Invalid time-ago format: `{}`. Must be on format <integer>(s|m|h|d|w)-ago or 'now'. E.g. '3d-ago' or '1w-ago'.".format(
                time_ago_string
            )
        )
    return ms


def get_fixed_freq(start_time, end_time, granularity):
    """Create a Timeseriesindex with a fixed frequency.

    Args:
        start_time (Union(pandas.Datetime,str, int, float)): Start time value to be converted to Timestamp.
        end_time (Union(pandas.Datetime,str, int, float)): End time value to be converted to Timestamp.
        granularity (str): Offset which Timestamp will have. Must be in format <integer>(s|m|h|d). E.g. '5m', '3h' or '1d'.".

    Returns:
        pd.DatetimeIndex: DatetimeIndex with fixed frequency.
    """
    return pd.date_range(start_time, end_time, freq=pd.Timedelta(milliseconds=granularity_to_ms(granularity)))


def above_below(series: pd.Series, upper_limit: float, lower_limit: float):
    """Above/below
    Count the number of points above/below upper_limit/lower_limit.

    Args:
        series (pandas.Series): Data with a pandas.DatetimeIndex.
        upper_limit (float): Value for upper limit.
        lower_limit (float): Value for lower limit.

    Returns:
        dict: The numbers above/below upper_limit/lower_limit.
    """
    result = {}
    if upper_limit is not None:
        result["above_upper"] = (series > upper_limit).sum()

    if lower_limit is not None:
        result["below_lower"] = (series < lower_limit).sum()

    return result


def num_vals_in_boxes(series: pd.Series, num_boxes: int):
    """Count the number of points in each of the boxes between 0 and 100 %.

    Args:
        series (pandas.Series): Data with a pandas.DatetimeIndex.
        num_boxes (int): Number of boxes.

    Returns:
        dict: A list with the number of points in each box in keyword "data".
    """
    max_val = series.max()
    min_val = series.min()
    delta = (max_val - min_val) / float(num_boxes)
    eps = 1.0e-6

    boxes = []
    low_limit = min_val
    for idx in range(num_boxes):
        if idx == 0:
            num_value_in_box = ((series >= low_limit * (1 - eps)) & (series <= low_limit + delta * (1 + eps))).sum()
        else:
            num_value_in_box = ((series > low_limit) & (series <= low_limit + delta * (1 + eps))).sum()
        boxes.append(num_value_in_box)
        low_limit += delta

    return {"data": boxes}


def functional_mean(function: Callable, x_vals: List) -> np.ndarray:
    """Convenience method to calculate the mean of a function inbetween each x value.
    The last point is the function called at the last x value. This is required for the
    input and output lists to have the same lengths.

    Args:
        function (Callable): Function used to calculate the mean.
        x_vals (List): Data values used to calculate the mean of a function.

    Raises:
        RuntimeError: Checks that x_vals is not empty.

    Returns:
        numpy.ndarray: Array with the calculated values.
    """

    # Catch empty input
    if len(x_vals) == 0:
        raise RuntimeError("No data in the input timeseries.")

    y_vals = np.zeros(len(x_vals))
    for i in range(len(x_vals) - 1):
        y_vals[i] = scipy.integrate.quad(function, x_vals[i], x_vals[i + 1])[0] / (x_vals[i + 1] - x_vals[i])
    y_vals[len(x_vals) - 1] = function(x_vals[-1])
    return y_vals


def is_na_all(data: Union[pd.DataFrame, pd.Series]) -> bool:
    """Convenience method to test if all values in dataframe/series are NaN.

    Args:
        data (Union[pandas.DataFrame, pandas.Series]): Input data.

    Raises:
        UserValueError: Checks that the data type for the input data is pandas.DataFrame or pandas.Series.

    Returns:
        bool: True if all values are Nan, False if all values are not Nan.
    """
    if isinstance(data, pd.Series):
        return data.isnull().all()
    elif isinstance(data, pd.DataFrame):
        return pd.isna(data).all(axis=None)
    else:
        raise UserValueError("Convenience method only supports Series or DataFrame.")


def gaps_detector(timestamps: np.ndarray, threshold: int = 86400) -> np.ndarray:
    """Detect gaps bigger than a threshold in time series
    Taken from: https://github.com/cognitedata/data-quality-functions/blob/master/functions/completeness/function/completeness.py

    Args:
        timestamps (numpy.ndarray): Time array_like.
        threshold (int, optional): Threshold for gaps. Defaults to 86400 seconds, i.e one full day in epoch time.
    Returns:
        numpy.ndarray: 2D array of begin and end time of gaps in seconds epoch time.
    """
    if len(timestamps) == 0:
        return None

    mask = np.diff(timestamps) >= threshold
    gaps = np.column_stack([timestamps[:-1][mask], timestamps[1:][mask]])

    return gaps


def check_uniform(data: pd.DataFrame, unit: str = "ns") -> bool:
    """Convenience method to verify input time series is uniform.
    Args:
        data (pandas.DataFrame): Time series to check. Object must have a datetime-like index.
        unit (str, optional): Time unit of uniformity check. Follows Numpy DateTime Units convention. Defaults to "ns".
    Returns:
        bool: True if input timeseries is uniform, False if not uniform.
    """
    # Get time difference between successive observations
    timediff = np.diff(data.index)
    # Round to desired time unit (e.g. uniform on second scale)
    floored = timediff.astype(dtype=f"timedelta64[{unit}]")
    # Check whether uniform
    return np.all(floored[0] == floored[1:])


def make_uniform(data: pd.DataFrame, resolution: str = "1s", interpolation: str = None) -> pd.DataFrame:
    """Convenience method to make an input time series dataframe uniform per specified temporal resolution.
    Object must have a datetime-like index.
    Args:
        resolution (str):  Temporal resolution of uniform time series. Follows Pandas DateTime convention. Defaults to "1s".
        interpolation (str, optional): Method for optional interpolation. Follows pandas Resampler.interpolate. Defaults to None.
    Returns:
        pandas.DataFrame: Uniform Timeseries DataFrame.
    """
    # If no interpolation specified, do not perform interpolation
    if not interpolation:
        data = data.resample(resolution).mean()
    # Otherwise perform desired interpolation
    else:
        data = data.resample(rule=resolution).mean().interpolate(method=interpolation)
    return data


def time_parse(time_window: str = "60min", function_name: str = "") -> pd.Timedelta:
    """Check that time_window is given in a correct format.

    Args:
        time_window (str, optional): Time window input. Defaults to "60min".
            Accepted string format: '3w', '10d', '5h', '30min', '10s'.
            If the user gives a number without unit (such as '60'), it will be considered as the number of minutes.
        function_name (str, optional): Prints the name of the function that called time_parse in the error message if an error is raised. Defaults to "".

    Raises:
        UserValueError: Raise a value error if time_window is given in a wrong format.

    Returns:
        str: parsed time_window in string format.
    """
    if str.isdigit(time_window):  # if user gives '60' it will be considered as '60min'
        warnings.warn(
            f"Missing time unit in argument 'time_window' in {function_name} function, assuming {time_window}min.",
            UserWarning,
        )
        time_window = str(time_window) + "min"
    try:
        return pd.Timedelta(time_window)
    except ValueError:
        raise UserValueError(
            f"Time window should be a string in weeks, days, hours or minutes format:'3w', '10d', '5h', '30min', '10s' not {time_window}."
        ) from None


def time_to_points(data: pd.Series, time_window="60min") -> int:
    """Find the average number of points given a time window.

    Args:
        data (pandas.Series): Data with a pandas.DatetimeIndex.
        time_window (str, optional): Length of the time period to compute the average of points. Defaults to "60min".
            Accepted string format: '3w', '10d', '5h', '30min', '10s'.

    Returns:
        int: average number of points in a time window.
    """
    time_window = time_parse(time_window)
    num_points = data.rolling(time_window).count().mean()

    return round(num_points)


def fill_gaps(
    data: pd.Series,
    granularity: str = None,
    ffill_resolution: str = None,
    interpolate_resolution: str = None,
    method: str = "ffill",
) -> pd.Series:
    """Mask nan/fill gaps bases on the time size of missing data

    Args:
        data (pandas.Series): Time series to impute.
        granularity (str, optional): Temporal resolution of uniform time series, before resampling. Defaults to None.
        ffill_resolution (str, optional): Gaps smaller than this will be filled with method. Defaults to None.
        interpolate_resolution (str, optional): Gaps smaller than this will be interpolated, larger than this will be filled by noise. Defaults to None.
        method (str, optional): Method used to fill nan. Defaults to 'ffill', but ‘backfill’, ‘bfill’ and ‘pad’ can also be used.
    Returns:
        pandas.DataFrame or pd.Series: Uniform, time series without nan values.
    """

    if not data.isnull().any():
        # no need to proceed
        return data

    if method not in ("backfill", "bfill", "pad", "ffill"):
        warnings.warn("Method should be in ('backfill', 'bfill', 'pad', 'ffill'). Set to ffill.", RuntimeWarning)
        method = "ffill"

    granularity_timedelata = pd.Timedelta("0T")
    try:
        granularity_timedelata = pd.Timedelta(granularity)
    except ValueError:
        raise ValueError(f"Unable to convert {granularity} to time resolution", RuntimeWarning)

    small_gap = 0
    if ffill_resolution:
        try:
            small_gap = granularity_timedelata / pd.Timedelta(ffill_resolution)
        except (ValueError, ZeroDivisionError):
            raise ValueError(f"Unable to convert {ffill_resolution} to time resolution", RuntimeWarning)

    inter_gap = None
    mask_gap = 0
    if interpolate_resolution is not None:
        try:
            mask_gap = granularity_timedelata / pd.Timedelta(interpolate_resolution)
            inter_gap = int(mask_gap)
        except (ValueError, ZeroDivisionError):
            raise ValueError(f"Unable to convert {interpolate_resolution} to time resolution", RuntimeWarning)

    # ffill all, after interpolate and adding noise
    # create groups based on size of gap
    nan_cumsum = (
        data.isnull().astype("int64").groupby(data.notnull().astype("int64").cumsum()).cumsum().mask(lambda x: x == 0)
    )
    helper = data.notnull()
    nan_group_size = nan_cumsum.groupby(helper.cumsum()).transform("max").mask(helper)

    # start with random filling for the big gaps
    mask = nan_group_size > inter_gap
    n = len(data[mask])
    data[mask] = np.random.normal(loc=np.mean(data), scale=np.std(data), size=n)

    # interpolate the inteprolate range
    mask = np.logical_and(nan_group_size > small_gap, nan_group_size <= mask_gap)
    data[mask] = data.interpolate(method="linear", limit_direction="forward", limit=inter_gap)

    return data.fillna(method=method)
